/*
 * Copyright (c) 2011-2014 gnome-pomodoro contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

.extension-pomodoro-indicator .system-status-icon {
    padding: 0;  /* default theme enlarges allocation with padding, which doesn't play nice with the DrawingArea */
    margin: 0 5px;
}

.extension-pomodoro-indicator-menu {
    max-width: 400px;
}

.extension-pomodoro-indicator-menu .state-item.active {
    font-weight: bold;
}

.extension-pomodoro-indicator-menu .button:insensitive {
    background-color: transparent;
}

.extension-pomodoro-indicator-menu .popup-menu-item > StLabel:first-child {
    padding-left: 16px;
}

.extension-pomodoro-toggle-menu-item {
    color: inherit;
    margin-top: 5px;
    margin-bottom: 3px;
    padding-top: 0;
    padding-bottom: 0;
}

.extension-pomodoro-toggle-menu-item .icon-button {
    margin-left: 5px;
}

.extension-pomodoro-timer-menu-item {
    color: inherit;
    margin-top: 5px;
    margin-bottom: 3px;
    padding: 0 10px 0 16px;
    spacing: 0;
}

.extension-pomodoro-timer-menu-item .extension-pomodoro-timer-button {
    color: inherit !important;
    padding: 0 5px;
    border-radius: 15px;
}

.extension-pomodoro-timer-menu-item .extension-pomodoro-timer-label {
    font-size: 1.6em;
    font-weight: 500;
    margin: 0 5px;
}

.extension-pomodoro-timer-menu-item .extension-pomodoro-timer-buttons-box {
    margin-left: 10px;
    spacing: 4px;
}

.extension-pomodoro-timer-label {
    margin: 0;
    padding: 0;
}

.extension-pomodoro-lightbox {
    background-color: rgba(0, 0, 0, 0.6);
}
.extension-pomodoro-lightbox-blurred {
    background-color: transparent;
}

.extension-pomodoro-overlay {
    color: white;
}

.extension-pomodoro-overlay .extension-pomodoro-overlay-container {
    margin: 30px;
}

.extension-pomodoro-overlay .extension-pomodoro-overlay-buttons {
    spacing: 10px;
}

.extension-pomodoro-overlay .icon-button {
    padding: 14px;
    background-color: rgba(255, 255, 255, 0.0);
}

.extension-pomodoro-overlay .icon-button:hover {
    background-color: rgba(255, 255, 255, 0.25);
}

.extension-pomodoro-overlay .icon-button:active {
    background-color: rgba(255, 255, 255, 0.3);
}

.extension-pomodoro-overlay .extension-pomodoro-overlay-contents {
    spacing: 15px;
}

.extension-pomodoro-overlay .extension-pomodoro-timer-label {
    font-size: 70pt;
    font-weight: 400;
}

.extension-pomodoro-overlay .extension-pomodoro-overlay-description {
    font-size: 14pt;
    font-weight: normal;
}

.extension-pomodoro-overlay .extension-pomodoro-do-not-touch-icon {
    margin-top: 20px;
    margin-bottom: 50px;
}

.extension-pomodoro-widget {
    font-size: 12pt;
    font-weight: normal;
    text-align: left;
    padding: 10px;
    border-radius: 15px;
    background-color: rgba(255, 255, 255, 0.06);
}

.extension-pomodoro-widget .extension-pomodoro-widget-content {
    spacing: 5px;
    margin: 5px;
    margin-left: 10px;
}

.extension-pomodoro-widget .extension-pomodoro-widget-title {
    font-size: 14pt;
    font-weight: bold;
}

.extension-pomodoro-widget .icon-button {
    border-radius: 10px;
    background-color: rgba(255, 255, 255, 0.0);  /* override gradient */
}

.extension-pomodoro-widget .icon-button.flat:hover {
    background-color: rgba(255, 255, 255, 0.1);  /* override gradient */
}

.extension-pomodoro-widget .icon-button.flat:active {
    background-color: rgba(255, 255, 255, 0.12);  /* override gradient */
}
